﻿Option Strict On

Imports Microsoft.CSharp
Imports System.Reflection
Imports System.IO
Imports System.Text
Imports System.Text.RegularExpressions

Public Class frmMain

#Region "Generator Setup"
    Private m_Parser As Parser = Nothing
    Private m_Working As Boolean = False

    Private Const WorkingDir As String = "Coco"
    Private Const FrameDirectory As String = "C#"
    Private Const CocoExecutable As String = "C#\coco.exe"

    Private WithEvents Generator As CocoRParserGenerator = New CSharpParserGenerator(CocoExecutable, WorkingDir, FrameDirectory)

    Private Sub frmMain_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        With txtGrammar
            .SelectionLength = 0
            .SelectionStart = txtGrammar.Text.Length

            .AddPattern("COMPILER|CHARACTERS|TOKENS|PRODUCTIONS|END", Color.Blue, True, False)
            .AddPattern("<.+?>", Color.Blue, False, False)
            .AddPattern("\(\..+?\.\)", Color.DarkViolet, True, False)
            .AddPattern("""" & "[^" & """" & "]+?" & """", Color.Brown, False, False)
        End With

        Parser = Nothing
    End Sub
#End Region

#Region "GUI State Control"
    Private Property Working() As Boolean
        Get
            Return m_Working
        End Get
        Set(ByVal value As Boolean)
            m_Working = value
            Static Controls As Control() = {txtGrammar, txtInput}
            Static ToolsStrips As ToolStripItem() = {NeuToolStripMenuItem, AusgabeLöschenToolStripMenuItem, KompilierenToolStripMenuItem}

            For Each c In Controls
                c.Enabled = Not m_Working
            Next

            For Each t In ToolsStrips
                t.Enabled = Not m_Working
            Next
        End Set
    End Property

    Private Property Parser() As Parser
        Get
            Return m_Parser
        End Get
        Set(ByVal value As Parser)
            m_Parser = value
            txtInput.Enabled = Parser IsNot Nothing
        End Set
    End Property

    Private Sub SetState(ByVal State As String)
        Me.lblState.Text = State
    End Sub

#End Region

#Region "GUI Events"
    Private Sub KompilierenToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles KompilierenToolStripMenuItem.Click
        Working = True
        Dim Grammar = txtGrammar.Text.Trim
        Generator.GenerateAsync(txtGrammar.Text)
    End Sub

    Private Sub AusgabeLöschenToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles AusgabeLöschenToolStripMenuItem.Click
        txtOutput.Text = ""
    End Sub

    Private Sub txtInput_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles txtInput.KeyDown
        If e.KeyCode = Keys.Return Then
            ParseOne()
            txtInput.Text = ""
        End If
    End Sub

    Private Sub NeuToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles NeuToolStripMenuItem.Click
        txtGrammar.Text = ""
        txtInput.Text = ""
        txtOutput.Text = ""
        Parser = Nothing
    End Sub
#End Region

#Region "Async Generator Events"

    Private Sub Generator_GenerationFailure(ByVal sender As Object, ByVal e As CocoRParserGenerator.GenerationFailureEventArgs) Handles Generator.GenerationFailure
        Me.Invoke(New Action(Of CocoRParserGenerator.GenerationFailureEventArgs)(AddressOf ShowFailure), e)
    End Sub

    Private Sub Generator_ProgessNotification(ByVal sender As Object, ByVal e As CocoRParserGenerator.ProgressNotifcationEventArgs) Handles Generator.ProgessNotification
        Me.Invoke(New Action(Of CocoRParserGenerator.ProgressNotifcationEventArgs)(AddressOf ShowProgress), e)
    End Sub

    Private Sub Generator_GenerationCompleted(ByVal sender As Object, ByVal e As CocoRParserGenerator.GenerationCompletedEventArgs) Handles Generator.GenerationCompleted
        Me.Invoke(New Action(Of CocoRParserGenerator.GenerationCompletedEventArgs)(AddressOf GenerationCompleted), e)
    End Sub

#End Region

#Region "Generator Events"
    Private Sub ShowFailure(ByVal e As CocoRParserGenerator.GenerationFailureEventArgs)
        Select Case e.Failure
            Case CocoRParserGenerator.Failure.InvalidGrammar
                Dim Errs = DirectCast(e.Data, IEnumerable(Of String)).Select(Function(s) " ** " & s).ToArray()
                Dim Err = String.Join(vbCrLf, Errs)
                MessageBox.Show("Grammatik fehlerhaft - Kein Parser geladen" & vbCrLf & Err, "Fehler", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)

            Case CocoRParserGenerator.Failure.InvalidParser
                Dim Errors = DirectCast(e.Data, IEnumerable(Of String)).Select(Function(s) " ** " & s).ToArray
                Dim Err = String.Join(vbCrLf, Errors)
                MessageBox.Show("Parser fehlerhaft - Kein Parser geladen" & vbCrLf & Err, "Fehler", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
        End Select

        Working = False
        Parser = Nothing
    End Sub

    Private Sub ShowProgress(ByVal e As CocoRParserGenerator.ProgressNotifcationEventArgs)
        Select Case e.Notification
            Case CocoRParserGenerator.Notification.Initializing
                SetState("Beginne Kompilation")

            Case CocoRParserGenerator.Notification.CompilingGrammar
                SetState("Kompiliere Grammatik")

            Case CocoRParserGenerator.Notification.CompilingParser
                SetState("Kompiliere Parser")
        End Select
    End Sub

    Private Sub GenerationCompleted(ByVal e As CocoRParserGenerator.GenerationCompletedEventArgs)
        SetState("Parser geladen")
        Working = False
        Parser = e.Parser
        txtInput.Focus()
    End Sub

#End Region

    Private Sub ParseOne()
        Dim Input = txtInput.Text

        Using Errs = New StringWriter
            Dim Result = Parser(Input, Errs)

            Dim Out As New StringBuilder
            Out.AppendLine("> " & Input)
            Out.AppendLine(Result)

            Dim Errors = Errs.ToString().Trim()
            If Errors <> "" Then Out.AppendLine("** " & Errors)

            Out.AppendLine()

            txtOutput.Text += Out.ToString()
        End Using

    End Sub

End Class
