﻿Imports System
Imports System.Collections.Generic
Imports System.Text
Imports System.Drawing
Imports System.Windows.Forms
Imports System.Text.RegularExpressions
Imports System.Runtime.InteropServices

' Basis-Implementierung dieser Klasse stammt von Fabian
' Siehe http://foren.activevb.de/cgi-bin/foren/view.pl?forum=21&msg=77&root=77
Class SyntaxRtb : Inherits RichTextBox
    ' Api-Deklarationen

    <DllImport("user32", CharSet:=CharSet.Auto)> _
    Private Shared Function SendMessage(ByVal hWnd As IntPtr, ByVal msg As Integer, ByVal wParam As Integer, ByVal lParam As Integer) As Integer
    End Function

    'Argumente für "SendMessage"
    Private Enum EditMessages
        LineIndex = 187
        LineFromChar = 201
        GetFirstVisibleLine = 206
        CharFromPos = 215
        PosFromChar = 1062
    End Enum

    Private Class RegexHighlightGroup
        Public Regex As Regex
        Public Font As Func(Of Font, Font)
        Public Group As Integer
        Public Color As Color
    End Class

    Private Const EM_SETEVENTMASK As Integer = 1073
    Private Const WM_SETREDRAW As Integer = 11

    Private updating As Integer
    Private oldEventMask As Integer

    ' Deaktiviert Redraw und Events der RichTextBox.
    Private Sub BeginUpdate(ByVal ctr As Control)

        ' Deal with nested calls.
        updating += 1
        If updating > 1 Then
            Exit Sub
        End If

        ' Prevent the control from raising any events.
        oldEventMask = SendMessage(ctr.Handle, EM_SETEVENTMASK, 0, 0)

        ' Prevent the control from redrawing itself.
        SendMessage(ctr.Handle, WM_SETREDRAW, 0, 0)
    End Sub

    ' Reaktiviert Redraw und Events der RichTextBox.
    Private Sub EndUpdate(ByVal ctr As Control)

        ' Deal with nested calls.
        updating -= 1
        If updating > 0 Then
            Exit Sub
        End If

        ' Allow the control to redraw itself.
        SendMessage(ctr.Handle, WM_SETREDRAW, 1, 0)

        ' Allow the control to raise event messages.
        SendMessage(ctr.Handle, EM_SETEVENTMASK, 0, oldEventMask)

        ctr.Invalidate()
    End Sub

    'Rückgängig Puffer
    Private UndoArgs As New List(Of UndoRedoArgs)()

    'Wiederholen Puffer
    Private RedoArgs As New List(Of UndoRedoArgs)()

    ' Speicher für Eigenschaften die beim Rückgängig machen und
    ' Wiederholen benötigt werden
    Private Class UndoRedoArgs
        Public Sub New(ByVal Text As String, ByVal selStart As Integer, ByVal selLength As Integer)
            Me.Text = Text
            Me.SelectionStart = selStart
            Me.SelectionLength = selLength
        End Sub

        Public Text As String
        Public SelectionStart As Integer
        Public SelectionLength As Integer
    End Class

    ' Gibt an, ob Rückgängig- oder Wiederholen-Aktionen
    ' durchgeführt werden
    Private UndoRedo As Boolean = False

    ' Text vor letzter Veränderung
    Private LastText As String = String.Empty

    ' Tab-Char
    Private tab As Char = vbTab(0)

    Private Matchings As New List(Of RegexHighlightGroup)

    Public Sub New()
        MyBase.New()
        Me.InitializeComponent()
    End Sub

    Private Sub InitializeComponent()
        Me.SuspendLayout()
        ' 
        ' SyntaxRtb
        ' 
        Me.AcceptsTab = True
        Me.BackColor = System.Drawing.Color.FromArgb(255, 255, 225)
        Me.EnableAutoDragDrop = True
        Me.Font = New System.Drawing.Font("Courier New", 9.75F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, CByte((0)))
        Me.ForeColor = System.Drawing.Color.Black

        Me.ResumeLayout(False)
    End Sub

    'Wenn der Text verändert wurde
    Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
        ' Wenn nicht rückgängig gemacht oder wiedergeholt wird
        If Not Me.UndoRedo Then
            ' Wiederholen Liste löschen
            Me.RedoArgs.Clear()

            Dim selStart As Integer = 0
            Dim selLength As Integer = 0
            Dim text As String = ""

            ' Wenn Text hinzugefügt wurde
            If Me.LastText.Length < Me.Text.Length Then
                selLength = Me.Text.Length - Me.LastText.Length
                selStart = Me.SelectionStart - selLength
            End If

            ' Wenn Text entfernt wurde
            If Me.LastText.Length > Me.Text.Length Then
                selLength = Me.LastText.Length - Me.Text.Length
                selStart = Me.SelectionStart
                text = Me.LastText.Substring(selStart, selLength)
            End If

            ' Text zur Rückgängig-Liste hinzufügen
            Me.UndoArgs.Add(New UndoRedoArgs(text, selStart, selLength))

            ' LastText mit aktuellem Text füllen
            Me.LastText = Me.Text
        End If
        ' UndoRedo-Aktionen beendet
        Me.UndoRedo = False

        ' Alle sichtbaren Zeilen formatieren
        Me.ColorVisibleLines()

        MyBase.OnTextChanged(e)
    End Sub

    'färbt alle sichtbaren Zeilen ein
    Private Sub ColorVisibleLines()
        ' Index der ersten sichtbaren Zeile
        Dim firstLine As Integer = Me.FirstVisibleLine
        ' Index der letzten sichtbaren Zeile
        Dim lastLine As Integer = Me.LastVisibleLine

        ' wenn einer der Indies kleiner als 0 ist...
        If (firstLine < 0) OrElse (lastLine < 0) Then
            ' abbrechen
            Exit Sub
        End If

        ' für jede sichtbare Zeile
        For i As Integer = firstLine To lastLine - 1
            ' Zeile syntaxhervorhebend färben
            Me.ColorLine(i)
        Next
    End Sub

    ' färbt die gesamte Rich Text Box ein
    Private Sub ColorRtb()
        For i As Integer = 0 To Me.Lines.Length - 1
            Me.ColorLine(i)
        Next
    End Sub

    'färbt eine Zeile syntaxhervorhebend ein
    Private Sub ColorLine(ByVal lineIndex As Integer)
        BeginUpdate(Me)

        ' start der aktuellen Zeile
        Dim start As Integer = Me.GetFirstCharIndex(lineIndex)

        ' aktuelle Zeile in Variable schreiben
        ' Tabs zu Leerzeichen umwandeln
        Dim line As String = Me.Lines(lineIndex).Replace(Convert.ToString(Me.tab), " ")

        ' alle Wörter
        Dim words As String() = line.Split(Convert.ToChar(" "))

        ' Position des aktuellen Wortes
        Dim wordStart As Integer = start

        'Speicher für Eigenschaften
        Dim selStart As Integer = Me.SelectionStart
        Dim selLength As Integer = Me.SelectionLength

        Dim selFont As Font = Me.Font
        Dim selColor As Color = Me.ForeColor

        'Zeile nach this.ForeColor und this.Font formatieren
        Me.SelectionStart = start
        Me.SelectionLength = line.Length

        Me.SelectionFont = selFont
        Me.SelectionColor = selColor

        For Each Highlighting In Matchings
            Dim Matches = Highlighting.Regex.Matches(line)

            For Each Match As Match In Matches
                Dim Word = Match.Groups(Highlighting.Group)
                Me.SelectionStart = start + Match.Index
                Me.SelectionLength = Word.Length
                Me.SelectionFont = Highlighting.Font(selFont)
                Me.SelectionColor = Highlighting.Color
            Next

        Next

        'Eigenschaften wiederherstellen
        Me.SelectionStart = selStart
        Me.SelectionLength = selLength
        Me.SelectionFont = selFont
        Me.SelectionColor = selColor

        EndUpdate(Me)
    End Sub

    'gibt die Nummer der ersten sichtbaren Zeile zurück
    Private ReadOnly Property FirstVisibleLine() As Integer
        Get
            Return SendMessage(Me.Handle, Convert.ToInt32(EditMessages.GetFirstVisibleLine), 0, 0)
        End Get
    End Property

    'gibt die Nummer der letzten sichtbaren Zeile zurück
    Private ReadOnly Property LastVisibleLine() As Integer
        Get
            Dim lastLine As Integer = Me.FirstVisibleLine + (Me.ClientSize.Height \ Me.Font.Height)

            If (lastLine > Me.Lines.Length) OrElse (lastLine = 0) Then
                Return Me.Lines.Length
            End If

            Return lastLine
        End Get
    End Property

    'gibt den Index des ersten sichtbaren Zeichens zurück
    Private ReadOnly Property FirstVisibleChar() As Integer
        Get
            Return Me.GetFirstCharIndex(Me.FirstVisibleLine)
        End Get
    End Property

    'gibt den Index des ersten Zeichens einer Zeile zurück
    Private Function GetFirstCharIndex(ByVal lineIndex As Integer) As Integer
        Return SendMessage(Me.Handle, Convert.ToInt32(EditMessages.LineIndex), lineIndex, 0)
    End Function

    Public Sub AddPattern(ByVal Pattern As String, ByVal FontColor As Color, ByVal Bold As Boolean, ByVal Italic As Boolean)
        AddPattern(Pattern, FontColor, Bold, Italic, 0)
    End Sub

    Public Sub AddPattern(ByVal Pattern As String, ByVal FontColor As Color, ByVal Bold As Boolean, ByVal Italic As Boolean, ByVal Group As Integer)
        Matchings.Add(New RegexHighlightGroup() With { _
                        .Regex = New Regex(Pattern, RegexOptions.Compiled), _
                        .Group = Group, _
                        .Color = FontColor, _
                        .Font = Function(Font) New Font(Font, _
                                                        If(Bold, FontStyle.Bold, FontStyle.Regular) Or _
                                                        If(Italic, FontStyle.Italic, FontStyle.Regular))})
        Me.ColorVisibleLines()
    End Sub

    'Einfügen
    'Paste-Methode überschreiben
    Public Shadows Sub Paste()
        'kein RTF sondern Text einfügen
        Me.SelectedText = Clipboard.GetText()
    End Sub

    'Kopieren
    'Copy-Methode überschreiben
    Public Shadows Sub Copy()
        'kein RTF sonder Text kopieren
        Clipboard.SetText(Me.SelectedText)
    End Sub

    'Protected Overrides Sub OnVScroll(ByVal e As System.EventArgs)
    '    Me.ColorVisibleLines()
    '    MyBase.OnVScroll(e)
    'End Sub

    ' Wenn eine Taste gedrückt wurde

    Protected Overloads Overrides Sub OnKeyDown(ByVal e As KeyEventArgs)
        If Me.ShortcutsEnabled Then
            ' Strg + V
            If (e.Control) AndAlso (e.KeyCode = Keys.V) Then
                ' einfügen
                Me.Paste()
                e.Handled = True
            End If

            ' Strg + C
            If (e.Control) AndAlso (e.KeyCode = Keys.C) Then
                ' kopieren
                Me.Copy()
                e.Handled = True
            End If

            ' Strg + Z
            If (e.Control) AndAlso (e.KeyCode = Keys.Z) Then
                ' Rückgängig
                Me.Undo()
                e.Handled = True
            End If

            ' Strg + V
            If (e.Control) AndAlso (e.KeyCode = Keys.Y) Then
                ' Wiederholen
                Me.Redo()
                e.Handled = True
            End If
        End If
        MyBase.OnKeyDown(e)
    End Sub

    'Schriftart wurde verändert
    Protected Overloads Overrides Sub OnFontChanged(ByVal e As EventArgs)
        'Gesamte Rich Text Box färben,
        'weil die berechnung auf der Schriftgröße aufbauen
        Me.ColorRtb()

        MyBase.OnFontChanged(e)
    End Sub

    'Vordergrundfarbe (Schriftfarbe) wurde verändert
    Protected Overloads Overrides Sub OnForeColorChanged(ByVal e As EventArgs)
        ' Sichtbare Zeilen einfärben
        Me.ColorVisibleLines()

        MyBase.OnForeColorChanged(e)
    End Sub

    ' Wenn die Größe verändert wurde
    Protected Overloads Overrides Sub OnSizeChanged(ByVal e As EventArgs)
        ' Sichtbare Zeilen einfärben
        Me.ColorVisibleLines()

        MyBase.OnSizeChanged(e)
    End Sub

    Public Shadows ReadOnly Property Rtf() As String
        Get
            'Wenn Rtf-Code abgefragt wird,
            'vorher gesamte Rich Text Box färben
            Me.ColorRtb()
            Return MyBase.Rtf
        End Get
    End Property

    ' Rückgängig machen
    Public Shadows Sub Undo()
        ' Wenn keine Elemente zum Rückgängig machen
        ' in der Liste sind: Abbrechen
        If Me.UndoArgs.Count = 0 Then
            Exit Sub
        End If

        ' OnTextChanged sagen, dass Rückgängig gemacht wird
        Me.UndoRedo = True

        ' Letztes Item der Rückgänig-Liste bestimmen
        Dim lastItem As Integer = Me.UndoArgs.Count - 1
        ' Letztes Item:
        Dim undoArg As UndoRedoArgs = UndoArgs(lastItem)

        ' Rückgänig machen durchführen
        Me.UnReDo(undoArg)

        ' Aktion zum Wiederholen hinzufügen
        Me.RedoArgs.Add(undoArg)
        ' Aktion löschen
        Me.UndoArgs.RemoveAt(lastItem)
    End Sub

    ' Wiederholen
    Public Shadows Sub Redo()
        ' Wenn keine Elemente zum Wiederholen
        ' in der Liste sind: Abbrechen
        If Me.RedoArgs.Count = 0 Then
            Exit Sub
        End If

        ' OnTextChanged sagen, dass wiedergeholt wird
        Me.UndoRedo = True

        ' Letztes Item der Wiederholen-Liste bestimmen
        Dim lastItem As Integer = Me.RedoArgs.Count - 1
        ' Letztes Item:
        Dim redoArg As UndoRedoArgs = RedoArgs(lastItem)

        ' Wiederholen durchführen
        Me.UnReDo(redoArg)

        ' Aktion zum Rückgängig machen hinzufügen
        Me.UndoArgs.Add(redoArg)
        ' Aktion löschen
        Me.RedoArgs.RemoveAt(lastItem)
    End Sub

    ' Rückgängig / Wiederholen
    ' durchführen
    Private Sub UnReDo(ByVal arg As UndoRedoArgs)
        Try
            ' Wenn Text leer
            If arg.Text = "" Then
                Me.SelectionStart = arg.SelectionStart
                Me.SelectionLength = arg.SelectionLength

                arg.Text = Me.SelectedText
                Me.SelectedText = ""
            Else
                ' Wenn Text nicht leer
                Me.SelectionStart = arg.SelectionStart
                Me.SelectedText = arg.Text

                arg.Text = ""
            End If
        Catch ex As Exception

        End Try
    End Sub

End Class